// Copyright 2015 The appc Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package tarheader

import (
	"archive/tar"
	"os"
)

var populateHeaderStat []func(h *tar.Header, fi os.FileInfo, seen map[uint64]string)

func Populate(h *tar.Header, fi os.FileInfo, seen map[uint64]string) {
	for _, pop := range populateHeaderStat {
		pop(h, fi, seen)
	}
}
